from sys import stdout
from decimal import Decimal
import numpy as np

# Mark Tillotson, 2017-07-19
# IIR Digital filter compiling to PASM

from utils import factorize_poly, quantize


coeff_bits = 18

PROP_XTAL = 5.000e6
PROP_CLK = PROP_XTAL * 16
CYCLE_TIME_US = 1.0e6 / PROP_CLK

I2S_XTAL = 6.144e6
I2S_CLK = I2S_XTAL * 16
I2S_CYCLE_TIME_US = 1.0e6 / I2S_CLK


def cycles_desc (cycles):
    return "%i cycles inc call (I2S: %s us, standard: %s us) %i instructs" % (cycles,
                                                                              Decimal (cycles * I2S_CYCLE_TIME_US).quantize (Decimal ('0.001')),
                                                                              Decimal (cycles * CYCLE_TIME_US).quantize (Decimal ('0.001')),
                                                                              cycles//4)

def gen_coeffs (s, ab, n, coeff, file = stdout):
    neg = coeff < 0.0
    if neg:
        coeff = -coeff
    icoeff = int (round (coeff * (1 << coeff_bits)))
    if icoeff == 2 << coeff_bits:
        icoeff -= 1
    print ("coeff%i_%s%i       ' value += t * %f [ %s%x.%04x ]" % (s, ab, n,
                                                                   -coeff if neg else coeff,
                                                                   '-' if neg else '+',
                                                                   icoeff >> coeff_bits, icoeff & ((1 << coeff_bits) - 1)),
           file = file)
    b = 1 << coeff_bits
    cycles = 0
    if (icoeff & b) != 0:
        print ("                %s     value, t" % ("sub" if neg else "add"), file = file)
        cycles += 4
        icoeff -= b
        
    c = 0
    while b != 0:
        while (icoeff & b) == 0:
            c += 1
            b >>= 1
            if b == 0:
                break
        if b == 0:
            break
        d = 0
        while (icoeff & b) != 0:
            d += 1
            b >>= 1

        if d == 1:
            print ("                sar     t, #%i" % c, file = file)
            print ("                %s     value, t" % ("sub" if neg else "add"), file = file)
            cycles += 8
        else:
            if c > 1:
                print ("                sar     t, #%i" % (c-1), file = file)
                cycles += 4
            print ("                %s     value, t" % ("sub" if neg else "add"), file = file)
            print ("                sar     t, #%i" % d, file = file)
            print ("                %s     value, t" % ("add" if neg else "sub"), file = file)
            cycles += 12
                
        c = 1

    cycles += 8
    print ("coeff%i_%s%i_ret   ret  ' %s" % (s, ab, n, cycles_desc (cycles)), file = file)
    print ("", file = file)
    return cycles



def gen_complex_coeffs (s, ab, coeff, file = stdout):
    print ("coeff%i_%s        ' v += t * %s" % (s, ab, coeff), file = file)
    print ("                mov     tx, x", file = file)
    print ("                mov     ty, y", file = file)
    cycles = 8
    cycles += gen_complex_coeffs_inner (s, coeff.real, file)

    print ("                neg     tx, y", file = file)
    print ("                mov     ty, x", file = file)
    cycles += 8
    cycles += gen_complex_coeffs_inner (s, coeff.imag, file)

    print ("coeff%i_%s_ret    ret" % (s, ab), file = file)
    print ("", file = file)
    return cycles + 4


def gen_complex_coeffs_inner (s, coeff, file):
    neg = coeff < 0.0
    if neg:
        coeff = -coeff
    icoeff = int (round (coeff * (1 << coeff_bits)))
    if icoeff == 2 << coeff_bits:
        icoeff -= 1

    b = 1 << coeff_bits
    cycles = 0
    if (icoeff & b) != 0:
        print ("                %s     vx, tx" % ("sub" if neg else "add"), file = file)
        print ("                %s     vy, ty" % ("sub" if neg else "add"), file = file)
        cycles += 8
        icoeff -= b
        
    c = 0
    while b != 0:
        while (icoeff & b) == 0:
            c += 1
            b >>= 1
            if b == 0:
                break
        if b == 0:
            break
        d = 0
        while (icoeff & b) != 0:
            d += 1
            b >>= 1

        if d == 1:
            print ("                sar     tx, #%i" % c, file = file)
            print ("                sar     ty, #%i" % c, file = file)
            print ("                %s     vx, tx" % ("sub" if neg else "add"), file = file)
            print ("                %s     vy, ty" % ("sub" if neg else "add"), file = file)
            cycles += 16
        else:
            if c > 1:
                print ("                sar     tx, #%i" % (c-1), file = file)
                print ("                sar     ty, #%i" % (c-1), file = file)
                cycles += 8
            print ("                %s     vx, tx" % ("sub" if neg else "add"), file = file)
            print ("                %s     vy, ty" % ("sub" if neg else "add"), file = file)
            print ("                sar     tx, #%i" % d, file = file)
            print ("                sar     ty, #%i" % d, file = file)
            print ("                %s     vx, tx" % ("add" if neg else "sub"), file = file)
            print ("                %s     vy, ty" % ("add" if neg else "sub"), file = file)
            cycles += 24
                
        c = 1

    return cycles



class Stage:
    def generate (self, file = stdout):
        raise Exception ("unimplemented")

    def process (self, sample):
        raise Exception ("unimplemented")

stage = 1


class OnePoleStage (Stage):
    def __init__ (self, z1, p1, bits = None):
        global stage
        assert z1.imag == 0 and p1.imag == 0
        self.z = [ z1 ]
        self.p = [ p1 ]
        self.b1 = quantize ((- p1).real, bits)
        self.a1 = quantize ((- z1).real, bits)
        self.w0 = 0.0
        self.w1 = 0.0
        self.stage = stage
        stage += 1

    def process (self, sample):
        self.w1 = self.w0
        self.w0 = sample - self.b1 * self.w1 
        return self.w0 + self.a1 * self.w1


    def generate (self, file = stdout):
        simple = isclose (abs (self.a1), 1.0)
        if self.b1.imag != 0:
            raise ValueError ("Not a real pole")
        print ("\n{{ --------------------------------------------- }}", file = file)
        print ("{{ Filter stage, one pole, z = %s, p = %s }}\n" % (self.z, self.p), file = file)
        cycles = gen_coeffs (self.stage, 'b', 1, self.b1.real, file)
        if not simple:
            cycles += gen_coeffs (self.stage, 'a', 1, self.a1.real, file)
        print ("stage%i" % self.stage, file = file)
        print ("                neg     t, w%i1" % self.stage, file = file)
        print ("                call    #coeff%i_b1  ' value -= w%i1 * %s" % (self.stage, self.stage, self.b1.real), file = file)
        print ("                mov     w%i0, value  ' w%i0 = value" % (2 * (self.stage,)), file = file)
        if simple:
            if self.a1.real < 0.0:
                print ("                sub     value, w%i1  ' value = w%i0 + w%i1" % (3 * (self.stage,)), file = file)
            else:
                print ("                add     value, w%i1  ' value = w%i0 + w%i1" % (3 * (self.stage,)), file = file)
        else:
            print ("                mov     t, w%i1" % self.stage, file = file)
            print ("                call    #coeff%i_a1  ' value += w%i1 * %s" % (self.stage, self.stage, self.ba.real), file = file)
            
        print ("                mov     w%i1, w%i0  ' w%i1 <- w%i0" % (4 * (self.stage,)), file = file)
        print ("stage%i_ret      ret" % self.stage, file = file)
        cycles += 4 * 6
        print ("", file = file)

        print ("w%i0             long    0" % self.stage, file = file)
        print ("w%i1             long    0" % self.stage, file = file)
        return cycles


class OnePoleComplexStage (Stage):
    def __init__ (self, z1, p1, bits = None):  # bits not used yet
        global stage
        self.z = [ z1 ]
        self.p = [ p1 ]
        self.b1 = - (quantize (p1.real, bits) + quantize(p1.imag, bits) * 1j)
        self.a1 = - (quantize (z1.real, bits) + quantize(z1.imag, bits) * 1j)
        self.w0 = 0.0
        self.w1 = 0.0
        self.stage = stage
        stage += 1

    def process (self, sample):
        self.w1 = self.w0
        self.w0 = sample - self.b1 * self.w1 
        return self.w0 + self.a1 * self.w1

    def generate (self, file = stdout):
        print ("\n{{ --------------------------------------------- }}", file = file)
        print ("{{ Filter stage, one complex pole, z = %s, p = %s }}\n" % (self.z, self.p), file = file)
        cycles = gen_complex_coeffs (self.stage, 'b', self.b1, file = file)
        cycles += gen_complex_coeffs (self.stage, 'a', self.a1, file = file)
        print ("stage%i" % self.stage, file = file)
        print ("                neg     x, w%i1x" % self.stage, file = file)
        print ("                neg     y, w%i1y" % self.stage, file = file)
        print ("                call    #coeff%i_b  ' value -= w%i1 * %s" % (self.stage, self.stage, self.b1), file = file)
        print ("                mov     w%i0x, vx  ' w%i0 = value" % (2 * (self.stage,)), file = file)
        print ("                mov     w%i0y, vy" % self.stage, file = file)
        print ("                mov     x, w%i1x" % self.stage, file = file)
        print ("                mov     y, w%i1y" % self.stage, file = file)
        print ("                call    #coeff%i_a  ' value += w%i1 * %s" % (self.stage, self.stage, self.a1), file = file)
            
        print ("                mov     w%i1x, w%i0x  ' w%i1 <- w%i0" % (4 * (self.stage,)), file = file)
        print ("                mov     w%i1y, w%i0y" % (2 * (self.stage,)), file = file)
        print ("stage%i_ret      ret" % self.stage, file = file)
        cycles += 4 * 11
        print ("", file = file)

        print ("w%i0x            long    0" % self.stage, file = file)
        print ("w%i0y            long    0" % self.stage, file = file)
        print ("w%i1x            long    0" % self.stage, file = file)
        print ("w%i1y            long    0" % self.stage, file = file)
        return cycles

class TwoPoleStage (Stage):
    def __init__ (self, z1, z2, p1, p2, bits = None):
        global stage
        assert z1.imag + z2.imag == 0
        assert p1.imag + p2.imag == 0
        
        #print ("Z1 %s" % z1)
        #print ("Z2 %s" % z2)
        self.z = [z1, z2]
        self.p = [p1, p2]
        self.b1 = quantize ((- p1 - p2).real, bits)
        self.b2 = quantize ((p1 * p2).real, bits)
        self.a1 = quantize ((- z1 - z2).real, bits)
        self.a2 = quantize ((z1 * z2).real, bits)
        self.w0 = 0.0
        self.w1 = 0.0
        self.w2 = 0.0
        self.stage = stage
        stage += 1


    def process (self, sample):
        self.w2 = self.w1
        self.w1 = self.w0
        self.w0 = sample - self.b1 * self.w1 - self.b2 * self.w2
        return self.w0 + self.a1 * self.w1 + self.a2 * self.w2


    def generate (self, file = stdout):
        simple = isclose (abs (self.a1), 2.0) and isclose (self.a2, 1.0)
        print ("\n{{ --------------------------------------------- }}\n", file = file)
        print ("{{ Filter stage, two pole, z = %s, p = %s }}\n" % (self.z, self.p), file = file)
        cycles =  gen_coeffs (self.stage, 'b', 1, self.b1.real, file)
        cycles += gen_coeffs (self.stage, 'b', 2, self.b2.real, file)
        if not simple:
            cycles += gen_coeffs (self.stage, 'a', 1, self.a1.real, file)
            cycles += gen_coeffs (self.stage, 'a', 2, self.a2.real, file)
        print ("stage%i" % self.stage, file = file)
        print ("                neg     t, w%i1" % self.stage, file = file)
        print ("                call    #coeff%i_b1  ' value -= w%i1 * %s" % (self.stage, self.stage, self.b1.real), file = file)
        print ("                neg     t, w%i2" % self.stage, file = file)
        print ("                call    #coeff%i_b2  ' value -= w%i2 * %s" % (self.stage, self.stage, self.b2.real), file = file)
        print ("                mov     w%i0, value  ' w%i0 = value" % (2 * (self.stage,)), file = file)
        if simple:
            if self.a1.real < 0.0:
                print ("                sub     value, w%i1" % self.stage, file = file)
                print ("                sub     value, w%i1" % self.stage, file = file)
            else:
                print ("                add     value, w%i1" % self.stage, file = file)
                print ("                add     value, w%i1" % self.stage, file = file)
            print ("                add     value, w%i2  ' value = w%i0 + 2 * w%i1 + w%i2" % (4 * (self.stage,)), file = file)
        else:
            print ("                mov     t, w%i1" % self.stage, file = file)
            print ("                call    #coeff%i_a1  ' value += w%i1 * %s" % (self.stage, self.stage, self.a1.real), file = file)
            print ("                mov     t, w%i2" % self.stage, file = file)
            print ("                call    #coeff%i_a2  ' value += w%i2 * %s" % (self.stage, self.stage, self.a2.real), file = file)
        print ("                mov     w%i2, w%i1" % (self.stage, self.stage), file = file)
        print ("                mov     w%i1, w%i0  ' w%i2 <- w%i1 <- w%i0" % (5 * (self.stage,)), file = file)
        print ("stage%i_ret      ret" % self.stage, file = file)
        print ("", file = file)
        cycles += 4 * 10

        print ("w%i0             long    0" % self.stage, file = file)
        print ("w%i1             long    0" % self.stage, file = file)
        print ("w%i2             long    0" % self.stage, file = file)
        return cycles




def compile_propeller_filter (resp, spinfile, bits):
    raw_zeroes = resp.zeros
    raw_poles = resp.poles
    raw_gain = resp.gain

    conj_zeroes, sing_zeroes, compl_zeroes = factorize_poly (raw_zeroes)
    conj_poles,  sing_poles,  compl_poles =  factorize_poly (raw_poles)

    conj_zeroes = sorted (conj_zeroes, key = lambda x: np.real(x))
    conj_poles  = sorted (conj_poles,  key = lambda x: -abs(x))

    print ("conj_zeroes %s" % conj_zeroes)
    print ("conj_poles %s" % [ abs(p) for p in conj_poles])

    complete_order_2_count = min (len (conj_zeroes), len (conj_poles))
    complete_order_1_count = min (len (sing_zeroes), len (sing_poles))
    zero_2_pole_1_count = min (len (conj_zeroes) - complete_order_2_count, len (sing_poles) - complete_order_1_count)
    zero_1_pole_2_count = min (len (conj_poles) - complete_order_2_count, len (sing_zeroes) - complete_order_1_count)
    zero_2_count = len (conj_zeroes) - complete_order_2_count - zero_2_pole_1_count
    pole_2_count = len (conj_poles) - complete_order_2_count - zero_1_pole_2_count
    zero_1_count = len (sing_zeroes) - complete_order_1_count - zero_1_pole_2_count
    pole_1_count = len (sing_poles) - complete_order_1_count - zero_2_pole_1_count

    print ("order2 = %s, order1 = %s" % (complete_order_2_count, complete_order_1_count))
    print ("zero_2_pole_1_count = %s" % zero_2_pole_1_count)
    print ("zero_1_pole_2_count = %s" % zero_1_pole_2_count)
    print ("zero_2_count = %s, zero_1_count = %s" % (zero_2_count, zero_1_count))
    print ("pole_2_count = %s, pole_1_count = %s" % (pole_2_count, pole_1_count))
    
    with open (spinfile, 'w') as f:
        stages = []
        if zero_1_count == 0 and zero_2_count == 0 and pole_1_count == 0 and pole_2_count == 0:
            if zero_2_pole_1_count == 0 and zero_1_pole_2_count == 0:
                for i in range (complete_order_2_count):
                    stages.append (TwoPoleStage (conj_zeroes[i], np.conj (conj_zeroes[i]),
                                                 conj_poles[i], np.conj (conj_poles[i]),
                                                 bits = bits))
                for i in range (complete_order_1_count):
                    stages.append (OnePoleStage (sing_zeroes[i],
                                                 sing_poles[i],
                                                 bits = bits))

        for i in range (max (len (compl_poles), len (compl_zeroes))):
            stages.append (OnePoleComplexStage (compl_zeroes[i],
                                                compl_poles[i],
                                                bits = bits))
        cycles = 0
        for stage in stages:
            cycles += stage.generate (f)
        print ("' %s bits per coefficient" % bits, file = f)
        print ("' %s" % cycles_desc (cycles), file = f)

"""
////////////////////////////////////////////////////////////////////////////////////////////
//                                TERMS OF USE: MIT License
////////////////////////////////////////////////////////////////////////////////////////////
// Permission is hereby granted, free of charge, to any person obtaining a copy of this 
// software and associated documentation files (the "Software"), to deal in the Software 
// without restriction, including without limitation the rights to use, copy, modify, merge,
// publish, distribute, sublicense, and/or sell copies of the Software, and to permit 
// persons to whom the Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or
// substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
// PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
// FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
// DEALINGS IN THE SOFTWARE.
////////////////////////////////////////////////////////////////////////////////////////////
"""
